# python

from django.db import models


class AccessLog(models.Model):
    id = models.AutoField(primary_key=True)
    timestamp = models.DateTimeField(auto_now_add=True)
    request_url = models.CharField(max_length=200)
    request_headers = models.JSONField()
    request_data = models.TextField()
    request_method = models.CharField(max_length=10)
    response_headers = models.JSONField()
    response_data = models.TextField()
    response_status_code = models.IntegerField()
    ip_address = models.GenericIPAddressField(null=True, blank=True)
    ipv4_address = models.GenericIPAddressField(null=True, blank=True)
    ipv6_address = models.GenericIPAddressField(null=True, blank=True)
    user_agent = models.CharField(max_length=255, null=True, blank=True)
    referer = models.CharField(max_length=255, null=True, blank=True)
    session_data = models.JSONField(null=True, blank=True)
    cookies = models.JSONField(null=True, blank=True)
    query_params = models.JSONField(null=True, blank=True)
    request_path = models.CharField(max_length=255, null=True, blank=True)
    request_protocol = models.CharField(max_length=10, null=True, blank=True)
    request_encoding = models.CharField(max_length=50, null=True, blank=True)
    response_time = models.FloatField(null=True, blank=True)
    request_body_size = models.IntegerField(null=True, blank=True)
    response_body_size = models.IntegerField(null=True, blank=True)
    response_is_ok = models.BooleanField(default=False)

    def __str__(self):
        return f'AccessLog {self.id}'



DEFAULT_DJ_ACCESS_LOGGER = {
    "method": "file",
    "separated_logging_db": False,
    "obfuscate_secrets": True,
    "logging_db_parameters": {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'logging_db.sqlite3'
    }
}



# python
from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.http import HttpResponse
from django.test import TestCase, RequestFactory
from django.utils.timezone import now

from dj_access_logger.middleware import AccessLogMiddleware


class AccessLogMiddlewareTests(TestCase):
    def setUp(self):
        self.factory = RequestFactory()
        self.middleware = AccessLogMiddleware(get_response=lambda r: HttpResponse("Test response"))
        User = get_user_model()
        self.user = User.objects.create_user(email='testuser@example.com', phone='1234567890', password='12345')

    def test_middleware_processes_request(self):
        request = self.factory.get('/test-url/')
        request.user = self.user

        self.middleware.process_request(request)
        self.assertIsNotNone(request.start_time)

    def test_middleware_processes_response(self):
        request = self.factory.get('/test-url/')
        request.user = self.user
        request.start_time = now()
        response = HttpResponse("Test response")

        with patch('dj_access_logger.repositories.repositories.LogData.log') as mock_log:
            processed_response = self.middleware.process_response(request, response)

            mock_log.assert_called_once()
            self.assertEqual(processed_response, response)

    def test_middleware_logs_post_request(self):
        post_data = {'key': 'value'}
        request = self.factory.post('/test-url/', data=post_data)
        request.user = self.user
        request.start_time = now()
        response = HttpResponse("Test response")

        with patch('dj_access_logger.repositories.repositories.LogData.log') as mock_log:
            self.middleware.process_response(request, response)

            mock_log.assert_called_once()
            call_args = mock_log.call_args[0][0]
            self.assertEqual(call_args.request.method, 'POST')
            self.assertEqual(call_args.request.request_path, '/test-url/')
            self.assertIn('key=value', call_args.request.data)

    def test_middleware_logs_ip_address(self):
        request = self.factory.get('/test-url/')
        request.user = self.user
        request.start_time = now()
        request.META['REMOTE_ADDR'] = '127.0.0.1'
        response = HttpResponse("Test response")

        with patch('dj_access_logger.repositories.repositories.LogData.log') as mock_log:
            self.middleware.process_response(request, response)

            mock_log.assert_called_once()
            call_args = mock_log.call_args[0][0]
            self.assertEqual(call_args.request.ip_address, '127.0.0.1')
            self.assertEqual(call_args.request.ipv4_address, '127.0.0.1')

    def test_middleware_logs_response_data(self):
        request = self.factory.get('/test-url/')
        request.user = self.user
        request.start_time = now()
        response = HttpResponse("Test response", status=201)

        with patch('dj_access_logger.repositories.repositories.LogData.log') as mock_log:
            self.middleware.process_response(request, response)

            mock_log.assert_called_once()
            call_args = mock_log.call_args[0][0]
            self.assertEqual(call_args.response.status_code, 201)
            self.assertEqual(call_args.response.data, "Test response")

    def test_get_client_ip(self):
        request = self.factory.get('/test-url/')
        request.META['HTTP_X_FORWARDED_FOR'] = '192.168.1.1, 10.0.0.1'

        ip, ipv4, ipv6 = self.middleware.get_client_ip(request)
        self.assertEqual(ip, '192.168.1.1')
        self.assertEqual(ipv4, '192.168.1.1')
        self.assertIsNone(ipv6)

        request.META['HTTP_X_FORWARDED_FOR'] = '2001:db8::1'
        ip, ipv4, ipv6 = self.middleware.get_client_ip(request)
        self.assertEqual(ip, '2001:db8::1')
        self.assertIsNone(ipv4)
        self.assertEqual(ipv6, '2001:db8::1')



# Generated by Django 5.1.1 on 2024-09-29 14:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dj_access_logger', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='accesslog',
            name='ipv4_address',
            field=models.GenericIPAddressField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='accesslog',
            name='ipv6_address',
            field=models.GenericIPAddressField(blank=True, null=True),
        ),
    ]



from dataclasses import dataclass
from typing import Dict, Any

from dj_access_logger.loggers.file_logger import FileLogger
from dj_access_logger.loggers.nosql_logger import NoSQLLogger
from dj_access_logger.loggers.sql_logger import SQLLogger
from dj_access_logger.utils import AccessLoggerSetting


@dataclass
class RequestData:
    url: str
    headers: Dict[str, Any]
    data: str | Dict
    method: str
    ip_address: str
    ipv4_address: str
    ipv6_address: str
    user_agent: str
    referer: str
    session_data: Dict[str, Any]
    cookies: Dict[str, Any]
    query_params: Dict[str, Any]
    request_path: str
    request_protocol: str
    request_encoding: str
    request_body_size: int


@dataclass
class ResponseData:
    headers: Dict[str, Any]
    data: str | Dict
    status_code: int
    response_time: float
    response_body_size: int


@dataclass
class LogData:
    request: RequestData
    response: ResponseData

    @property
    def settings(self):
        return AccessLoggerSetting()

    def get_logger(self):
        if self.settings.method == 'sql':
            return SQLLogger()
        elif self.settings.method == 'file':
            return FileLogger()
        elif self.settings.method == 'nosql':
            return NoSQLLogger()

    def log(self):
        self.get_logger().log(self)



# python

from .abstract_logger import AbstractLogger


class NoSQLLogger(AbstractLogger):
    def log(self, log_data):
        # Implement the logic to log data to a NoSQL database
        # This is a placeholder implementation
        raise NotImplementedError("NoSQL logging is not yet implemented")



import json

from django.contrib import admin
from django.template.loader import render_to_string
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from pygments import highlight, lexers, formatters

from .models import AccessLog


@admin.register(AccessLog)
class AccessLogAdmin(admin.ModelAdmin):
    list_display = ('id', 'timestamp', 'request_url', 'request_method', 'response_status_code', 'ip_address')
    search_fields = ('request_headers', 'response_headers')
    list_filter = (
    'request_url', 'request_method', 'response_status_code', 'ip_address', 'ipv4_address', 'ipv6_address')
    readonly_fields = (
        'id', 'timestamp', 'request_url', 'request_headers', 'request_data', 'request_method', 'response_headers',
        'response_data', 'response_status_code', 'session_data'
    )

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def request_headers(self, obj):
        return format_html(self.render_copyable_field(obj.request_headers))

    def request_data(self, obj):
        return format_html(self.render_copyable_field(obj.request_data))

    def response_headers(self, obj):
        return format_html(self.render_copyable_field(obj.response_headers))

    def response_data(self, obj):
        return format_html(self.render_response_data(obj.response_data, obj.response_headers))

    def session_data(self, obj):
        return format_html(self.render_copyable_field(obj.session_data))

    @staticmethod
    def render_copyable_field(self, value):  # https://daniel.feldroy.com/posts/pretty-formatting-json-django-admin#
        """Function to display pretty version of our data"""
        # Convert the data to sorted, indented JSON
        response = json.dumps(value, sort_keys=True, indent=2)
        # Truncate the data. Alter as needed
        # response = response[:5000]
        # Get the Pygments formatter
        formatter = formatters.HtmlFormatter(style='staroffice', linenos='table')  # pylint: disable=no-member
        # Highlight the data
        response = highlight(response, lexers.JsonLexer(), formatter)  # pylint: disable=no-member
        # Get the stylesheet
        style = "<style>" + formatter.get_style_defs() + "</style><br>"

        copy_button = """
            <script>
                window.onCopyCodeButtonClick = function(element) {
                    let codeContainer = element.parentElement.getElementsByClassName("code")[0];
                    navigator.clipboard.writeText(codeContainer.innerText);
                }
            </script>
            <button type="button" onclick="onCopyCodeButtonClick(this)" title="Copy">Copy JSON</button>
        """

        # Safe the output
        return mark_safe(style + response + copy_button)

    def render_response_data(self, response_data, response_headers):
        content_type = response_headers.get('Content-Type', '')
        if 'application/json' in content_type:
            formatted_value = json.dumps(json.loads(response_data), indent=4)
            return render_to_string('admin/copyable_field.html', {'value': formatted_value})
        elif 'text/html' in content_type:
            return format_html(response_data)
        elif 'exception' in response_data.lower():
            return format_html('<pre>{}</pre>', response_data)
        else:
            return format_html('<div style="white-space: pre-wrap;">{}</div>', response_data)



from django.conf import settings

from .default_settings import DEFAULT_DJ_ACCESS_LOGGER


class AccessLoggerSetting:
    def __init__(self):
        user_settings = getattr(settings, 'DJ_ACCESS_LOGGER', {})
        self.method = user_settings.get("method", DEFAULT_DJ_ACCESS_LOGGER["method"])
        self.separated_logging_db = user_settings.get("separated_logging_db",
                                                      DEFAULT_DJ_ACCESS_LOGGER["separated_logging_db"])
        self.obfuscate_secrets = user_settings.get("obfuscate_secrets", DEFAULT_DJ_ACCESS_LOGGER["obfuscate_secrets"])
        self.logging_db_parameters = user_settings.get("logging_db_parameters",
                                                       DEFAULT_DJ_ACCESS_LOGGER["logging_db_parameters"])
        if self.separated_logging_db:
            self.set_separated_logging_db()
        self.validate()

    def set_separated_logging_db(self):
        settings.DATABASES['logging'] = self.logging_db_parameters
        if isinstance(settings.DATABASE_ROUTERS, tuple):
            settings.DATABASE_ROUTERS = list(settings.DATABASE_ROUTERS)
        if not isinstance(settings.DATABASE_ROUTERS, list):
            settings.DATABASE_ROUTERS = []

        router_path = 'dj_access_logger.db_router.LoggingDBRouter'
        settings.DATABASE_ROUTERS.append(router_path)

    def validate(self):
        if self.method not in ['file', 'sql', 'nosql']:
            raise ValueError("DJ_ACCESS_LOGGER['method'] must be 'file', 'sql' or 'nosql'")

        if self.method == 'nosql' and not self.logging_db_parameters:
            raise ValueError("DJ_ACCESS_LOGGER['logging_db_parameters'] must be set for 'nosql' method")


DJ_ACCESS_LOGGER_SETTINGS = AccessLoggerSetting()



# python
import json

from .abstract_logger import AbstractLogger


class FileLogger(AbstractLogger):
    def log(self, log_data):
        response_status_code = log_data.response.status_code
        response_is_ok = 200 <= response_status_code < 300
        log_entry = {
            'request_url': log_data.request.url,
            'request_headers': log_data.request.headers,
            'request_data': log_data.request.data,
            'request_method': log_data.request.method,
            'response_headers': log_data.response.headers,
            'response_data': log_data.response.data,
            'response_status_code': response_status_code,
            'ip_address': log_data.request.ip_address,
            'user_agent': log_data.request.user_agent,
            'referer': log_data.request.referer,
            'session_data': log_data.request.session_data,
            'cookies': log_data.request.cookies,
            'query_params': log_data.request.query_params,
            'request_path': log_data.request.request_path,
            'request_protocol': log_data.request.request_protocol,
            'request_encoding': log_data.request.request_encoding,
            'response_time': log_data.response.response_time,
            'request_body_size': log_data.request.data,
            'response_body_size': log_data.response.response_body_size,
            'response_is_ok': response_is_ok
        }

        with open('access_log.json', 'a') as log_file:
            log_file.write(json.dumps(log_entry) + '\n')



# dj_access_logger/management/commands/setup_dj_access_logger.py

from django.apps import apps
from django.core.management.base import BaseCommand
from django.db import connections
from django.db.migrations.executor import MigrationExecutor


class Command(BaseCommand):
    help = 'Set up the logging database for dj_access_logger'

    def handle(self, *args, **options):
        self.stdout.write("Setting up logging database...")

        # Get the logging database connection
        connection = connections['logging']

        # Check if there are any tables in the database
        with connection.cursor() as cursor:
            cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
            tables = cursor.fetchall()

        if not tables:
            self.stdout.write("No tables found. Creating tables...")

            # Get the app's migration operations
            app_config = apps.get_app_config('dj_access_logger')
            migration_executor = MigrationExecutor(connection)
            targets = [(app_config.label, None)]

            # Run the migrations
            migration_executor.migrate(targets)

            self.stdout.write(self.style.SUCCESS("Successfully created tables for dj_access_logger."))
        else:
            self.stdout.write("Tables already exist in the logging database.")

        self.stdout.write(self.style.SUCCESS("Logging database setup complete."))



# python
import ipaddress

from django.utils.deprecation import MiddlewareMixin
from django.utils.timezone import now
from django.http import FileResponse

from dj_access_logger.repositories.repositories import RequestData, ResponseData, LogData


class AccessLogMiddleware(MiddlewareMixin):
    def process_request(self, request):
        request.start_time = now()
        # Read and store the request body to avoid multiple reads
        request._body = request.body if request.body else b''

    def process_response(self, request, response):
        try:
            request_body = request._body.decode('utf-8') if request._body else ''
        except UnicodeDecodeError:
            request_body = "<Binary Data>"

        if isinstance(response, FileResponse):
            response_data_content = "<File Response>"
        else:
            response_data_content = response.content.decode('utf-8') if response.content else ''

        request_data = RequestData(
            url=request.build_absolute_uri(),
            headers=dict(request.headers),
            data=request_body,
            method=request.method,
            ip_address=self.get_client_ip(request)[0],
            ipv4_address=self.get_client_ip(request)[1],
            ipv6_address=self.get_client_ip(request)[2],
            user_agent=request.META.get('HTTP_USER_AGENT', ''),
            referer=request.META.get('HTTP_REFERER', ''),
            session_data=dict(request.session.items()),
            cookies=request.COOKIES,
            query_params=request.GET,
            request_path=request.path,
            request_protocol=request.scheme,
            request_encoding=request.encoding,
            request_body_size=len(request_body)
        )

        response_data = ResponseData(
            headers=dict(response.items()),
            data=response_data_content,
            status_code=response.status_code,
            response_time=(now() - request.start_time).total_seconds(),
            response_body_size=len(response_data_content),
        )

        LogData(request=request_data, response=response_data).log()

        return response

    @staticmethod
    def get_client_ip(request):
        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            ip = x_forwarded_for.split(',')[0]
        else:
            ip = request.META.get('REMOTE_ADDR')

        ipv4_address = None
        ipv6_address = None

        try:
            ip_obj = ipaddress.ip_address(ip)
            if ip_obj.version == 4:
                ipv4_address = ip
            elif ip_obj.version == 6:
                ipv6_address = ip
                if ip_obj.ipv4_mapped:
                    ipv4_address = str(ip_obj.ipv4_mapped)
        except ValueError:
            pass

        return ip, ipv4_address, ipv6_address



# Generated by Django 5.1.1 on 2024-09-29 12:24

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AccessLog',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('timestamp', models.DateTimeField(auto_now_add=True)),
                ('request_url', models.CharField(max_length=200)),
                ('request_headers', models.JSONField()),
                ('request_data', models.TextField()),
                ('request_method', models.CharField(max_length=10)),
                ('response_headers', models.JSONField()),
                ('response_data', models.TextField()),
                ('response_status_code', models.IntegerField()),
                ('ip_address', models.GenericIPAddressField(blank=True, null=True)),
                ('user_agent', models.CharField(blank=True, max_length=255, null=True)),
                ('referer', models.CharField(blank=True, max_length=255, null=True)),
                ('session_data', models.JSONField(blank=True, null=True)),
                ('cookies', models.JSONField(blank=True, null=True)),
                ('query_params', models.JSONField(blank=True, null=True)),
                ('request_path', models.CharField(blank=True, max_length=255, null=True)),
                ('request_protocol', models.CharField(blank=True, max_length=10, null=True)),
                ('request_encoding', models.CharField(blank=True, max_length=50, null=True)),
                ('response_time', models.FloatField(blank=True, null=True)),
                ('request_body_size', models.IntegerField(blank=True, null=True)),
                ('response_body_size', models.IntegerField(blank=True, null=True)),
                ('response_is_ok', models.BooleanField(default=False)),
            ],
        ),
    ]



# dj_access_logger/db_router.py

class LoggingDBRouter:
    """
    A router to control all database operations on models in the
    dj_access_logger application.
    """

    def db_for_read(self, model, **hints):
        if model._meta.app_label == 'dj_access_logger':
            return 'logging'
        return None

    def db_for_write(self, model, **hints):
        if model._meta.app_label == 'dj_access_logger':
            return 'logging'
        return None

    def allow_relation(self, obj1, obj2, **hints):
        if obj1._meta.app_label == 'dj_access_logger' or obj2._meta.app_label == 'dj_access_logger':
            return True
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        if app_label == 'dj_access_logger':
            return db == 'logging'
        elif db == 'logging':
            return app_label == 'dj_access_logger'
        return None

    def allow_syncdb(self, db, model):
        if model._meta.app_label == 'dj_access_logger':
            return db == 'logging'
        elif db == 'logging':
            return model._meta.app_label == 'dj_access_logger'
        return None

    def connection_allows_schema_migrate(self, connection, app_label):
        if app_label == 'dj_access_logger':
            return connection.alias == 'logging'
        return None



class MethodAccessLogger:
    def __init__(self, logger=None):
        self.logger = logger or logging.getLogger(__name__)

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self.logger.info(f"Called {func.__name__}")
            return func(*args, **kwargs)

        return wrapper

    def __get__(self, instance, owner):
        return self if instance is None else types.MethodType(self, instance)



from setuptools import setup, find_packages

setup(
    name='dj-access-logger',
    version='1.0.5',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',
    description='A Django middleware for logging HTTP requests and responses.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/elkhayyat/dj-access-logger',
    author='AHMED ELKHAYYAT',
    author_email='elkhayyat.me@gmail.com',
    install_requires=[
        'Django>=3.0',
        'pymongo',
        'djongo',
        'mysqlclient',
        'psycopg2',
        'pygments'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



# python

from dj_access_logger.models import AccessLog
from .abstract_logger import AbstractLogger


class SQLLogger(AbstractLogger):
    def log(self, log_data):
        response_status_code = log_data.response.status_code
        response_is_ok = 200 <= response_status_code < 300
        AccessLog.objects.create(
            request_url=log_data.request.url,
            request_headers=log_data.request.headers,
            request_data=log_data.request.data,
            request_method=log_data.request.method,
            response_headers=log_data.response.headers,
            response_data=log_data.response.data,
            response_status_code=response_status_code,
            ip_address=log_data.request.ip_address,
            user_agent=log_data.request.user_agent,
            referer=log_data.request.referer,
            session_data=log_data.request.session_data,
            cookies=log_data.request.cookies,
            query_params=log_data.request.query_params,
            request_path=log_data.request.request_path,
            request_protocol=log_data.request.request_protocol,
            request_encoding=log_data.request.request_encoding,
            response_time=log_data.response.response_time,
            request_body_size=len(log_data.request.data),
            response_body_size=log_data.response.response_body_size,
            response_is_ok=response_is_ok
        )



# python
# dj_access_logger/apps.py

import os

from django.apps import AppConfig
from django.core.management import call_command

from .utils import DJ_ACCESS_LOGGER_SETTINGS


class DjAccessLoggerConfig(AppConfig):
    name = 'dj_access_logger'

    def ready(self):
        if DJ_ACCESS_LOGGER_SETTINGS.separated_logging_db:
            DJ_ACCESS_LOGGER_SETTINGS.set_separated_logging_db()
            if DJ_ACCESS_LOGGER_SETTINGS.logging_db_parameters.get('ENGINE') == 'django.db.backends.sqlite3':
                db_path = DJ_ACCESS_LOGGER_SETTINGS.logging_db_parameters.get('NAME', 'logging_db.sqlite3')
                if not os.path.exists(db_path):
                    print(f"Creating logging sqlite database at {db_path}")
                    open(db_path, 'w').close()
